print("import: 'axelrod'")
import axelrod

print("import: 'axelrod.data'")
import axelrod.data

print("import: 'axelrod.strategies'")
import axelrod.strategies

print("import: 'axelrod.tests'")
import axelrod.tests

